/* $File: //depot/main/ticWorkspaceGadget/src/WgInputServerFilter.cpp $
 * $Date: 2004/09/11 $
 * $Revision: #15 $
 *
 * (C) 1999-2004 by Mikael Jansson <apps@mikael.jansson.be>
 */
#include "WgInputServerFilter.h"

#include <Message.h>
#include <Point.h>
#include <InterfaceDefs.h>
#include <Region.h>
#include <Roster.h>
#include <Window.h>
#include <Beep.h>
#include <cstdio>
#include <math.h>

#include <List.h>

#define WC_WIDTH	0
#define WC_HEIGHT	1

static uint8 g_WorkspaceConfig[32][2] = {
{1,1},	// 1
{2,1},	// 2
{3,1},	// 3
{2,2},	// 4
{5,1},	// 5
{3,2},	// 6
{7,1},	// 7
{4,2},	// 8
{3,3}, 	// 9
{5,2}, 	// 10
{11,1},	// 11
{4,3},	// 12
{13,1},	// 13
{7,2},	// 14
{5,3},	// 15
{4,4},	// 16
{17,1},	// 17
{6,3},	// 18
{19,1},	// 19
{5,4},	// 20
{7,3},	// 21
{11,2},	// 22
{23,1},	// 23
{6,4},	// 24
{5,5},	// 25
{13,2},	// 26
{9,3},	// 27
{7,4},	// 28
{29,1},	// 29
{6,6},	// 30
{31,1},	// 31
{8,4}	// 32
};


WgInputServerFilter::WgInputServerFilter() :
	BInputServerFilter(),
	m_WarpingEnabled(T_DEF_WARPING_ENABLED),
	m_WarpingDelay(T_DEF_WARPING_DELAY),
	m_PopupEnabled(T_DEF_POPUP_ENABLED),
	m_PopupAnimationEnabled(T_DEF_POPUP_ANIMATION_ENABLED),
	m_PopupSize(T_DEF_POPUP_SIZE),
	m_PopupPosition(T_DEF_POPUP_POSITION),
	m_PopupDirection(T_DEF_POPUP_DIRECTION),
	m_CurrentWorkspace(33),
	m_LastWorkspace(33),
	m_LastTime(0),
	m_MouseSpeed(0),
	m_AppRunning(true),
	m_CommPort(0),
	m_PopupMsgr(0),
	m_ConfigThread(0)
{
	m_LastPoint.x = m_LastPoint.y = -1;
	for (int32 i=0; i<T_MOUSE_SPEEDS; i++)
		m_LastPoints[i].x = m_LastPoints[i].y = m_LastSpeeds[i] = 0.0;
		
	if (m_ConfigThread == 0) {
		m_ConfigThread = spawn_thread(config_thread_function, "ticWorkspaceGadget one", 1, this);
	}
	resume_thread(m_ConfigThread);

	LoadSettings();
}

WgInputServerFilter::~WgInputServerFilter() {

	delete_port(m_CommPort);

	m_AppRunning = false;

	kill_thread(m_ConfigThread);
	delete g_AppPrefs;
}
	
filter_result WgInputServerFilter::Filter(BMessage *p_Msg, BList *p_List) {

	// process the message
	switch(p_Msg->what) {
		case B_MOUSE_MOVED: {
		
			// get the screen size
			BRegion region;
			GetScreenRegion(&region);
			m_ScreenFrame = region.Frame();

			RefreshStateIfNeeded();

			BPoint point;
			int32 buttons=0;
			bigtime_t when;
			
			p_Msg->FindInt64("when", &when);
			p_Msg->FindPoint("where", &point);
			p_Msg->FindInt32("buttons", &buttons);
			
			UpdateSpeed(point, when);
			
			// good mouse speed threshold: 0.7
			
			if (m_PopupEnabled && MouseAtCorner(point))
				SetPopupHidden(false);
			else if (MouseOutsidePopup(point))
				SetPopupHidden(true);
			
			// else (?)
			ChangeWorkspaceIfAttacked(point);
			

		} break;
	}

	return B_DISPATCH_MESSAGE;
}

status_t WgInputServerFilter::InitCheck() {
	return BInputServerFilter::InitCheck();
}

// input_server loading.

BInputServerFilter* instantiate_input_filter() {
	return new WgInputServerFilter();
}

// P R I V A T E   F U N C T I O N S ///////////////////////////////////////////

void WgInputServerFilter::ChangeWorkspaceIfAttacked(BPoint p_Point) {
	// Note: takes the direction into account.
	// If the mouse is moving in another direction than
	// towards the edge, nothing will happen.
	// Also, the mouse speed needs to be higher than a certain threshold.

	enum {
		Left=0,Up=1,Right=2,Down=3,NoDirection=4
	};

	uint8 direction = NoDirection;
	if (p_Point.x == m_ScreenFrame.right && m_MouseDirection == T_MOUSE_DIRECTION_HORIZONTAL)
		direction = Right;
	else if (p_Point.x == m_ScreenFrame.left && m_MouseDirection == T_MOUSE_DIRECTION_HORIZONTAL)
		direction = Left;
	else if (p_Point.y == m_ScreenFrame.top && m_MouseDirection == T_MOUSE_DIRECTION_VERTICAL)
		direction = Up;
	else if (p_Point.y == m_ScreenFrame.bottom && m_MouseDirection == T_MOUSE_DIRECTION_VERTICAL)
		direction = Down;

	float threshold = m_WarpingDelay/100.0;
	
	BRect rect = PopupFrame();
	// shrink rect
	if (m_PopupPosition == T_POS_TOP_LEFT) {
		rect.bottom -= m_PopupSize/2;
		rect.right -= m_PopupSize/2;
	} else if (m_PopupPosition == T_POS_TOP_RIGHT) {
		rect.left += m_PopupSize/2;
		rect.bottom -= m_PopupSize/2;
	} else if (m_PopupPosition == T_POS_BOTTOM_RIGHT) {
		rect.top += m_PopupSize/2;
		rect.left += m_PopupSize/2;
	} else if (m_PopupPosition == T_POS_BOTTOM_LEFT) {
		rect.top += m_PopupSize/2;
		rect.right -= m_PopupSize/2;
	}
	
	if (m_MouseSpeed > threshold && direction < NoDirection && !rect.Contains(p_Point)) {
			
			uint8 height = g_WorkspaceConfig[count_workspaces()-1][WC_HEIGHT];
			uint8 width = g_WorkspaceConfig[count_workspaces()-1][WC_WIDTH];

			// Remember: m_CurrentWorkspace is 0-based!
			// row and column will also be 0-based.
			
			// the formula goes something like this:
			uint8 row = m_CurrentWorkspace / width;
			uint8 column = m_CurrentWorkspace % width;
			
			// example: current workspace = #7 = 6, count = 16, width/height=4
			// then: 	row = #2 = 1
			// 			column = #3 = 2
			if (direction == Right) {
				if (column < width-1)
					column++;
			} else if (direction == Left) {
				if (column > 0)
					column--;
			} else if (direction == Down) {
				if (row < height-1)
					row++;
			} else if (direction == Up) {
				if (row > 0)
					row--;
			}
			
			uint32 wspace = row*width + column;
			if (wspace != m_CurrentWorkspace) {
				m_CurrentWorkspace = wspace;
				m_MouseSpeed = 0.0;
				for (int32 i=0; i<T_MOUSE_SPEEDS; i++) {
					m_LastSpeeds[i] = 0.0;
					m_LastPoints[i].x = m_LastPoints[i].y = 0.0;
				}
				
				activate_workspace(m_CurrentWorkspace);
				RefreshStateIfNeeded();
			}
	}

}

//------------------------------------------------------------------------------

void WgInputServerFilter::UpdateSpeed(BPoint p_Point, bigtime_t p_When) {

	double distance = sqrt((m_LastPoint.x-p_Point.x)*(m_LastPoint.x-p_Point.x) +
						(m_LastPoint.y-p_Point.y)*(m_LastPoint.y-p_Point.y));

	if (distance > 0) {
		m_LastSpeeds[m_CurrentSample] = distance / (double)(p_When - m_LastTime)*1000;
		m_LastPoints[m_CurrentSample] = p_Point;

		// average speed
		double total = 0.000001;
		for (int32 i=0; i<T_MOUSE_SPEEDS; i++)
			total += m_LastSpeeds[i];
		m_MouseSpeed = total/T_MOUSE_SPEEDS;
		
		// get average direction
		int first_point = (m_CurrentSample + 5) % (T_MOUSE_SPEEDS-1);
		float slope_x = m_LastPoints[first_point].x - p_Point.x;
		float slope_y = m_LastPoints[first_point].y - p_Point.y;
		if (slope_x < 0.0) slope_x *= -1.0;
		if (slope_y < 0.0) slope_y *= -1.0;

		m_MouseDirection = (slope_x > slope_y) ? T_MOUSE_DIRECTION_HORIZONTAL : T_MOUSE_DIRECTION_VERTICAL;
		
		m_LastTime = p_When;
		m_LastPoint = p_Point;
		m_CurrentSample = (m_CurrentSample+1) % (T_MOUSE_SPEEDS-1);
	}
}

//------------------------------------------------------------------------------

bool WgInputServerFilter::MouseOutsidePopup(BPoint p_Point) {
	return !PopupFrame().Contains(p_Point);
	
}

//------------------------------------------------------------------------------

bool WgInputServerFilter::MouseAtCorner(BPoint p_Point) {
	bool result = false;
	// top left, top right, bottom right, bottom left

	if ((m_PopupPosition == T_POS_TOP_LEFT && p_Point.x == m_ScreenFrame.left && p_Point.y == m_ScreenFrame.top) ||
		(m_PopupPosition == T_POS_TOP_RIGHT && p_Point.x == m_ScreenFrame.right && p_Point.y == m_ScreenFrame.top) ||
		(m_PopupPosition == T_POS_BOTTOM_RIGHT && p_Point.x == m_ScreenFrame.right && p_Point.y == m_ScreenFrame.bottom) ||
		(m_PopupPosition == T_POS_BOTTOM_LEFT && p_Point.x == m_ScreenFrame.left && p_Point.y == m_ScreenFrame.bottom))
		result = true;
	return result;
}

//------------------------------------------------------------------------------

void WgInputServerFilter::LoadSettings() {

	if (g_AppPrefs == NULL) {
		g_AppPrefs = new SettingsFile("Settings", "tic_software/ticWorkspaceGadget");
	}
	g_AppPrefs->Load();

	RefreshStateIfNeeded();
}

//------------------------------------------------------------------------------

int32 WgInputServerFilter::config_thread_function(void *p_Data) {
	return static_cast<WgInputServerFilter*>(p_Data)->ConfigThread();
}

//------------------------------------------------------------------------------

int32 WgInputServerFilter::ConfigThread() {

	m_CommPort = find_port(kCommunicationPortName);
	if (m_CommPort == B_NAME_NOT_FOUND) {
		// ok, the app hasn't yet been started.
		// we'll create the port ourselves.
		m_CommPort = create_port(10, kCommunicationPortName);
	}

	while (m_AppRunning) {
		int32 id,data;
		ssize_t size = read_port(m_CommPort, &id, (void*)&data,	sizeof(data));
		if (size == sizeof(data)) {
			LoadSettings();
		}
	}
	
	return B_OK;
}

//------------------------------------------------------------------------------

BRect WgInputServerFilter::PopupFrame() {

	StartPopupIfNeeded();

	BMessage msg(B_GET_PROPERTY), reply;
	msg.AddSpecifier("Frame");
	msg.AddSpecifier("Window", 0L);	
	if (m_PopupMsgr && m_PopupMsgr->IsValid())
		m_PopupMsgr->SendMessage(&msg, &reply);
	
	BRect rect;
	reply.FindRect("result", &rect);

	return rect;
}

//------------------------------------------------------------------------------

void WgInputServerFilter::RefreshStateIfNeeded() {
	m_LastWorkspace = m_CurrentWorkspace;
	m_CurrentWorkspace = current_workspace();

	uint8 old_direction = m_PopupDirection;
	uint8 old_position = m_PopupPosition;
	uint16 old_size = m_PopupSize;

	m_WarpingEnabled = T_DEF_WARPING_ENABLED;
	m_WarpingDelay = T_DEF_WARPING_DELAY;
	m_PopupEnabled = T_DEF_POPUP_ENABLED;
	m_PopupAnimationEnabled = T_DEF_POPUP_ANIMATION_ENABLED;
	m_PopupSize = T_DEF_POPUP_SIZE;
	m_PopupPosition = T_DEF_POPUP_POSITION;
	m_PopupDirection = T_DEF_POPUP_DIRECTION;

	int32 index = 0;
	BMessage msg;
	int32 workspace_index = 0;
	bool found = false;
	while (g_AppPrefs->FindMessage(kPrefsWarpingMsg, index++, &msg) == B_OK) {
		msg.FindInt32(kPrefsWarpingIndexField, &workspace_index);
		if (workspace_index == (int32)m_CurrentWorkspace) {
			found = true;
			break;
		}
	}

	if (found) {
		// enabled
		bool enabled;
		if (msg.FindBool(kPrefsPopupEnabled, &enabled) == B_OK) {
			m_PopupEnabled = enabled;
		}

		// animation enabled
		if (msg.FindBool(kPrefsPopupAnimationEnabled, &enabled) == B_OK) {
			m_PopupAnimationEnabled = enabled;
		}

		// size
		uint16 size;
		if (msg.FindInt16(kPrefsPopupSize, (int16*)&size) == B_OK) {
			m_PopupSize = size;
		}
		
		// position
		uint8 position;
		if (msg.FindInt8(kPrefsPopupPosition, (int8*)&position) == B_OK) {
			m_PopupPosition = position;
		}
		
		// direction
		uint8 direction;
		if (msg.FindInt8(kPrefsPopupDirection, (int8*)&direction) == B_OK) {
			m_PopupDirection = direction;
		}
	}

	bool needs_refresh = false;
	float x,y;
	if (m_PopupEnabled) {

		// horizontal direction
		if (m_PopupDirection == T_DIR_HORIZONTAL) {
			if (m_PopupPosition == T_POS_TOP_LEFT) {
				x = m_ScreenFrame.left;
				y = m_ScreenFrame.top;
			} else
			if (m_PopupPosition == T_POS_TOP_RIGHT) {
				x = m_ScreenFrame.right-m_PopupSize;
				y = m_ScreenFrame.top;
			} else
			if (m_PopupPosition == T_POS_BOTTOM_RIGHT) {
				x = m_ScreenFrame.right-m_PopupSize;
				y = m_ScreenFrame.bottom-m_PopupSize;
			} else
			if (m_PopupPosition == T_POS_BOTTOM_LEFT) {
				x = m_ScreenFrame.left;
				y = m_ScreenFrame.bottom-m_PopupSize;
			}
		}
		// vertical direction
		else {
			if (m_PopupPosition == T_POS_TOP_LEFT) {
				x = m_ScreenFrame.left;
				y = m_ScreenFrame.top;
			} else
			if (m_PopupPosition == T_POS_TOP_RIGHT) {
				x = m_ScreenFrame.right-m_PopupSize;
				y = m_ScreenFrame.top;
			} else
			if (m_PopupPosition == T_POS_BOTTOM_RIGHT) {
				x = m_ScreenFrame.right-m_PopupSize;
				y = m_ScreenFrame.bottom-m_PopupSize;
			} else
			if (m_PopupPosition == T_POS_BOTTOM_LEFT) {
				x = m_ScreenFrame.left;
				y = m_ScreenFrame.bottom-m_PopupSize;
			}
		}

	
#if 0
		// horizontal direction
		if (m_PopupDirection == T_DIR_HORIZONTAL) {
			if (m_PopupPosition == T_POS_TOP_LEFT) {
				x = m_ScreenFrame.left - m_PopupSize-5.f;
				y = m_ScreenFrame.top;
			} else
			if (m_PopupPosition == T_POS_TOP_RIGHT) {
				x = m_ScreenFrame.right+5.f;
				y = m_ScreenFrame.top;
			} else
			if (m_PopupPosition == T_POS_BOTTOM_RIGHT) {
				x = m_ScreenFrame.right+5.f;
				y = m_ScreenFrame.bottom-m_PopupSize;
			} else
			if (m_PopupPosition == T_POS_BOTTOM_LEFT) {
				x = m_ScreenFrame.left-m_PopupSize-5.f;
				y = m_ScreenFrame.bottom-m_PopupSize;
			}
		}
		// vertical direction
		else {
			if (m_PopupPosition == T_POS_TOP_LEFT) {
				x = m_ScreenFrame.left;
				y = m_ScreenFrame.top-m_PopupSize-5.f;
			} else
			if (m_PopupPosition == T_POS_TOP_RIGHT) {
				x = m_ScreenFrame.right-m_PopupSize;
				y = m_ScreenFrame.top-m_PopupSize-5.f;
			} else
			if (m_PopupPosition == T_POS_BOTTOM_RIGHT) {
				x = m_ScreenFrame.right-m_PopupSize;
				y = m_ScreenFrame.bottom+m_PopupSize+5.f;
			} else
			if (m_PopupPosition == T_POS_BOTTOM_LEFT) {
				x = m_ScreenFrame.left;
				y = m_ScreenFrame.bottom+m_PopupSize+5.f;
			}
		}
#endif

		// Find the popup's position
		// Could've been handled nicer. :)
	
		/*
		BPoint point = CurrentPopupPosition();
		if (point.x < (m_ScreenFrame.right-m_ScreenFrame.left)/2 &&
			point.y < (m_ScreenFrame.bottom-m_ScreenFrame.top)/2) {
			old_position = T_POS_TOP_LEFT;
		} else
		if (point.x > (m_ScreenFrame.right-m_ScreenFrame.left)/2 &&
			point.y < (m_ScreenFrame.bottom-m_ScreenFrame.top)/2) {
			old_position = T_POS_TOP_RIGHT;
		} else
		if (point.x < (m_ScreenFrame.right-m_ScreenFrame.left)/2 &&
			point.y > (m_ScreenFrame.bottom-m_ScreenFrame.top)/2) {
			old_position = T_POS_BOTTOM_LEFT;
		} else
		if (point.x > (m_ScreenFrame.right-m_ScreenFrame.left)/2 &&
			point.y > (m_ScreenFrame.bottom-m_ScreenFrame.top)/2) {
			old_position = T_POS_BOTTOM_RIGHT;
		}
		*/
		

		if ((m_LastWorkspace != m_CurrentWorkspace) || (old_position != m_PopupPosition) || (old_direction != m_PopupDirection) || (old_size != m_PopupSize)) {
			needs_refresh = true;
		}
	}

	if (needs_refresh) {

		StartPopupIfNeeded();

		BMessage msg(B_SET_PROPERTY), reply;
	
		BRect rect(x,y, (float)(x+m_PopupSize), (float)(y+m_PopupSize));
		msg.MakeEmpty();
		msg.AddSpecifier("Frame");
		msg.AddSpecifier("Window", 0L);
		msg.AddRect("data", rect);
		if (m_PopupMsgr && m_PopupMsgr->IsValid())
			m_PopupMsgr->SendMessage(&msg, &reply);
	
		msg.MakeEmpty();
		msg.AddSpecifier("Flags");
		msg.AddSpecifier("Window", 0L);
		msg.AddInt32("data", (int32)(B_NOT_MOVABLE|B_NOT_RESIZABLE|B_WILL_ACCEPT_FIRST_CLICK));
		if (m_PopupMsgr && m_PopupMsgr->IsValid())
			m_PopupMsgr->SendMessage(&msg, &reply);
	
		msg.MakeEmpty();
		msg.AddSpecifier("Look");
		msg.AddSpecifier("Window", 0L);
		msg.AddInt32("data", (int32)(B_NO_BORDER_WINDOW_LOOK));
		if (m_PopupMsgr && m_PopupMsgr->IsValid())
			m_PopupMsgr->SendMessage(&msg, &reply);

	}

}

//------------------------------------------------------------------------------

void WgInputServerFilter::StartPopupIfNeeded() {
	// start popup, set flags/look, hide
	bool valid = true;
	if (m_PopupMsgr)
		if (!m_PopupMsgr->IsValid())
			valid = false;

	if (!m_PopupMsgr || !valid) {
		app_info app;
		be_roster->GetAppInfo(kPopupAppSig, &app);
			
		if (app.team == -1) {
			be_roster->Launch(kPopupAppSig);
		}
		m_PopupMsgr = new BMessenger(kPopupAppSig);
	}
}


//------------------------------------------------------------------------------

void WgInputServerFilter::SetPopupHidden(bool p_Hidden) {
	BMessage msg_set(B_SET_PROPERTY), reply;
	msg_set.AddSpecifier("Hidden");
	msg_set.AddSpecifier("Window", 0L);
	msg_set.AddBool("data", p_Hidden);
	if (m_PopupMsgr->IsValid())
		m_PopupMsgr->SendMessage(&msg_set, &reply);
}

//------------------------------------------------------------------------------

void WgInputServerFilter::PopupMoveTo(float p_X, float p_Y) {

	StartPopupIfNeeded();

	BMessage msg_set(B_SET_PROPERTY), reply;

	BMessage msg(B_GET_PROPERTY);
	msg.AddSpecifier("Frame");
	msg.AddSpecifier("Window", 0L);	
	if (m_PopupMsgr->IsValid())
		m_PopupMsgr->SendMessage(&msg, &reply);
	
	BRect rect;
	reply.FindRect("result", &rect);
	rect.OffsetTo(p_X, p_Y);
	
	msg.what = B_SET_PROPERTY;
	msg.MakeEmpty();
	msg.AddSpecifier("Frame");
	msg.AddSpecifier("Window", 0L);
	msg.AddRect("data", rect);
	if (m_PopupMsgr->IsValid())
		m_PopupMsgr->SendMessage(&msg, &reply);
}

