/* $File: //depot/main/ticWorkspaceGadget/src/WgPrefsWindow.cpp $
 * $Date: 2004/09/10 $
 * $Revision: #5 $
 *
 * (C) 1999-2004 by Mikael Jansson <apps@mikael.jansson.be>
 */
#include "WgPrefsWindow.h"
#include "WgGlobals.h"

#include <Application.h>
#include <PopUpMenu.h>
#include <MenuItem.h>
#include <Message.h>
#include <MenuField.h>

#include "layout/HGroup.h"
#include "layout/VGroup.h"
#include "layout/MBox.h"
#include "layout/MBorder.h"
#include "layout/Space.h"
#include "layout/MButton.h"

#include <cstdio>

enum {
	kCurrentWorkspace = 'kcws',
	kAllWorkspaces = 'kaws',
	kWarpingEnabled = 'kwpe',
	kPopupEnabled = 'kped',
	kPopupAnimationEnabled = 'kpae',
	kApplyButton = 'kapb',
};

WgPrefsWindow::WgPrefsWindow(BRect p_Rect) :
	MWindow(p_Rect, "ticWorkspaceGadget", B_TITLED_WINDOW,
		B_ASYNCHRONOUS_CONTROLS|B_NOT_ZOOMABLE|B_NOT_RESIZABLE),
	m_Workspaces(0),
	m_WarpingDelay(0),
	m_WarpingEnabled(0),
	m_PopupEnabled(0),
	m_PopupSize(0),
	m_PopupPosition(0),
	m_Mode(T_CURRENT_WORKSPACE),
	m_CommPort(0),
	m_WorkspaceMessageIndex(-1)
{
	
	SetWorkspaces(B_ALL_WORKSPACES);
	
	m_Workspaces = new BPopUpMenu("Current Workspace");
	m_Workspaces->AddItem(new BMenuItem("Current Workspace", new BMessage(kCurrentWorkspace)));
	m_Workspaces->AddItem(new BMenuItem("All Workspaces", new BMessage(kAllWorkspaces)));

	BMenuField *m_WorkspaceField = new BMenuField(BRect(10,10,10,10), "moo", NULL, m_Workspaces);
	m_WorkspaceField->SetDivider(0.0f);

	m_WarpingEnabled = new MCheckBox("Enabled", new BMessage(kWarpingEnabled));

	m_WarpingDelay = new SpinButton("Threshold", SPIN_INTEGER);
	m_WarpingDelay->SetMaximum(1500);
	m_WarpingDelay->SetStepSize(10.0);

	m_PopupEnabled = new MCheckBox("Enabled", new BMessage(kPopupEnabled));
	m_PopupAnimationEnabled = new MCheckBox("Animate", new BMessage(kPopupAnimationEnabled));

	m_PopupSize = new SpinButton("Size", SPIN_INTEGER);
	m_PopupSize->SetMaximum(1000);
	m_PopupSize->SetStepSize(10.0);

	m_PopupPosition = new MRadioGroup("Top left", "Top right", "Bottom left", "Bottom right", 0);

	m_PopupDirection = new MRadioGroup("Horizontal", "Vertical", 0);

	// load the settings from the current workspace
	// *only* the first time when we start the app
	LoadSettings();

	MView *content_view =
	new VGroup(
		new Space(minimax(5,5,5,5)),
		new HGroup(
			new Space(minimax(5,5,5,5)),
			new MBox(B_PLAIN_BORDER, "Warping",
				new VGroup(
					m_WarpingEnabled,
					m_WarpingDelay,
					0
				),
				minimax(-1,-1,-1,-1)
			),
			new Space(),
			0
		),
		new Space(minimax(5,5,5,5)),
		new MBox(B_PLAIN_BORDER, "Popup",
			new VGroup(
				new VGroup(
					m_PopupEnabled,
					//m_PopupAnimationEnabled,
					m_PopupSize,
					//new Space(minimax(5,5,5,5)),
					//new MBox(B_PLAIN_BORDER, "Position",
					//	new VGroup(
							m_PopupPosition,
					//		0
					//	)
					//),
					0
				),
				/*
				new Space(minimax(5,5,5,5)),
				new Space(),
				new MBox(B_PLAIN_BORDER, "Direction",
					new VGroup(
						m_PopupDirection,
						0
					),
					minimax(-1,-1,-1,-1)
				),
				*/
				0
			),
			minimax(-1,-1,-1,-1)
		),
		new Space(minimax(5,5,5,5)),
		new HGroup(
			new MButton("Set on All Workspaces", new BMessage(kApplyButton), NULL, minimax(-1,-1,-1,-1)),
			new Space(),
			0
		),
		0
	);

	MView *root = new VGroup(
		new MBorder(M_RAISED_BORDER,3,"blah",
			content_view
		),
		0
	);

	
	AddChild(dynamic_cast<BView*>(root));
}

//------------------------------------------------------------------------------

bool WgPrefsWindow::QuitRequested() {

	g_AppPrefs->ReplacePoint(kPrefsAppWindowSize,
		BPoint(Frame().left, Frame().top));

	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}

//------------------------------------------------------------------------------

void WgPrefsWindow::MessageReceived(BMessage *p_Msg) {

	switch (p_Msg->what) {

		case kApplyButton: {
			SaveSettingsAllWorkspaces();
		} break;

		case kWarpingEnabled: {
			SetWarpingEnabled(m_WarpingEnabled->Value());
		} break;
		
		case kPopupEnabled: {
			SetPopupEnabled(m_PopupEnabled->Value());
		} break;

		case kPopupAnimationEnabled: {
			SetPopupAnimationEnabled(m_PopupAnimationEnabled->Value());
		} break;
		
		case '!spn': {
			void *ptr;
			p_Msg->FindPointer("source", &ptr);
			SpinButton *button = dynamic_cast<SpinButton*>(static_cast<BControl*>(ptr));
			if (button == m_WarpingDelay) {
				SetWarpingDelay((int)button->Value());
			} else
			if (button == m_PopupSize) {
				SetPopupSize((int)button->Value());
			}
		} break;
		
		case '!rad': {
			void *ptr;
			p_Msg->FindPointer("radio", &ptr);
			MRadioGroup *radio = static_cast<MRadioGroup*>(ptr);
			if (radio == m_PopupPosition) {
				SetPopupPosition(radio->ActiveButton());
			} else
			if (radio == m_PopupDirection) {
				SetPopupDirection(radio->ActiveButton());
			}
		} break;
		
		default:
			MWindow::MessageReceived(p_Msg);
			break;
	
	}
}

//------------------------------------------------------------------------------

void WgPrefsWindow::WorkspacesChanged(uint32 p_NewMask, uint32 p_OldMask) {
}

void WgPrefsWindow::WorkspaceActivated(int32 p_Index, bool p_Enabled) {
	if (p_Enabled) {
		LoadSettings();
	}
}

// P R I V A T E   F U N C T I O N S ///////////////////////////////////////////

void WgPrefsWindow::SetWarpingEnabled(bool p_Enabled) {

	BMessage *msg = CurrentWorkspaceMessage();
	msg->ReplaceBool(kPrefsWarpingEnabled, p_Enabled);
	SaveCurrentWorkspaceMessage(msg);
	BroadcastChanges();
}

//------------------------------------------------------------------------------

void WgPrefsWindow::SetPopupEnabled(bool p_Enabled) {

	BMessage *msg = CurrentWorkspaceMessage();
	msg->ReplaceBool(kPrefsPopupEnabled, p_Enabled);
	SaveCurrentWorkspaceMessage(msg);
	BroadcastChanges();
}

//------------------------------------------------------------------------------

void WgPrefsWindow::SetPopupAnimationEnabled(bool p_Enabled) {

	BMessage *msg = CurrentWorkspaceMessage();
	msg->ReplaceBool(kPrefsPopupAnimationEnabled, p_Enabled);
	SaveCurrentWorkspaceMessage(msg);
	BroadcastChanges();
}

//------------------------------------------------------------------------------
	
void WgPrefsWindow::SetWarpingDelay(int p_Value) {

	BMessage *msg = CurrentWorkspaceMessage();
	msg->ReplaceInt16(kPrefsWarpingDelay, (int16)p_Value);
	SaveCurrentWorkspaceMessage(msg);
	BroadcastChanges();
}

//------------------------------------------------------------------------------

void WgPrefsWindow::SetPopupSize(int p_Value) {

	BMessage *msg = CurrentWorkspaceMessage();
	msg->ReplaceInt16(kPrefsPopupSize, (int16)p_Value);
	SaveCurrentWorkspaceMessage(msg);
	BroadcastChanges();
}

//------------------------------------------------------------------------------

void WgPrefsWindow::SetPopupPosition(int p_State) {

	BMessage *msg = CurrentWorkspaceMessage();
	msg->ReplaceInt8(kPrefsPopupPosition, (int8)p_State);
	SaveCurrentWorkspaceMessage(msg);
	BroadcastChanges();
}

//------------------------------------------------------------------------------

void WgPrefsWindow::SetPopupDirection(int p_State) {

	BMessage *msg = CurrentWorkspaceMessage();
	msg->ReplaceInt8(kPrefsPopupDirection, (int8)p_State);
	SaveCurrentWorkspaceMessage(msg);
	BroadcastChanges();
}

//------------------------------------------------------------------------------

void WgPrefsWindow::BroadcastChanges() {

	g_AppPrefs->Save();

	if (m_CommPort == 0) {
		// can't see a case where the filter won't keep up with
		// the data sent, but make the buffer 10 long just in case.
		m_CommPort = find_port(kCommunicationPortName);
		if (m_CommPort == B_NAME_NOT_FOUND)
			m_CommPort = create_port(10, kCommunicationPortName);
	}

	// tell the boys to reload
	int32 blah = 'blah';
	if (write_port_etc(m_CommPort, 'blah', &blah, sizeof(blah), B_TIMEOUT, 4000) == B_BAD_PORT_ID) {
		m_CommPort = find_port(kCommunicationPortName);
		if (m_CommPort == B_NAME_NOT_FOUND)
			m_CommPort = create_port(10, kCommunicationPortName);
		write_port(m_CommPort, 'blah', &blah, sizeof(blah));
	}
}

//------------------------------------------------------------------------------

void WgPrefsWindow::LoadSettings() {

	BMessage *msg;
	msg = CurrentWorkspaceMessage();

	bool boolean;
	uint16 integer16;
	uint8 integer8;

	msg->FindBool(kPrefsWarpingEnabled, &boolean);
	m_WarpingEnabled->SetValue(boolean);
	
	msg->FindInt16(kPrefsWarpingDelay, (int16*)&integer16);
	m_WarpingDelay->SetValue((long)integer16);
	
	msg->FindBool(kPrefsPopupEnabled, &boolean);
	m_PopupEnabled->SetValue(boolean);
	
	msg->FindBool(kPrefsPopupAnimationEnabled, &boolean);
	m_PopupAnimationEnabled->SetValue(boolean);
	
	msg->FindInt16(kPrefsPopupSize, (int16*)&integer16);
	m_PopupSize->SetValue((long)integer16);
	
	msg->FindInt8(kPrefsPopupPosition, (int8*)&integer8);
	m_PopupPosition->SetActive(integer8);
	
	msg->FindInt8(kPrefsPopupDirection, (int8*)&integer8);
	m_PopupDirection->SetActive(integer8);

	BroadcastChanges();
	
	delete msg;
}

//------------------------------------------------------------------------------

BMessage *WgPrefsWindow::CurrentWorkspaceMessage() {

	int32 workspace = current_workspace();
	int32 index = 0;
	BMessage msg;
	int32 workspace_index = 0;
	bool found = false;
	BMessage *settings;
	while (g_AppPrefs->FindMessage(kPrefsWarpingMsg, index, &msg) == B_OK) {
		msg.FindInt32(kPrefsWarpingIndexField, &workspace_index);
		if (workspace_index == workspace) {
			found = true;
			break;
		}
		index++;
	}

	m_WorkspaceMessageIndex = -1;

	if (found) {
		settings = new BMessage(msg);
		m_WorkspaceMessageIndex = index;
	} else {
		settings = new BMessage;

		settings->AddInt32(kPrefsWarpingIndexField, workspace);

		settings->AddBool(kPrefsWarpingEnabled, T_DEF_WARPING_ENABLED);
		settings->AddInt16(kPrefsWarpingDelay, T_DEF_WARPING_DELAY);
		settings->AddBool(kPrefsPopupEnabled, T_DEF_POPUP_ENABLED);
		settings->AddBool(kPrefsPopupAnimationEnabled, T_DEF_POPUP_ANIMATION_ENABLED);
		settings->AddInt16(kPrefsPopupSize, T_DEF_POPUP_SIZE);
		settings->AddInt8(kPrefsPopupPosition, T_DEF_POPUP_POSITION);
		settings->AddInt8(kPrefsPopupDirection, T_DEF_POPUP_DIRECTION);
	}
	
	return settings;
}

//------------------------------------------------------------------------------

void WgPrefsWindow::SaveCurrentWorkspaceMessage(BMessage *p_Msg) {

	if (m_WorkspaceMessageIndex >= 0) {
		g_AppPrefs->ReplaceMessage(kPrefsWarpingMsg, m_WorkspaceMessageIndex, p_Msg);
	} else {
		g_AppPrefs->AddMessage(kPrefsWarpingMsg, p_Msg);
	}
	

	delete p_Msg;
}

//------------------------------------------------------------------------------

void WgPrefsWindow::SaveSettingsAllWorkspaces() {

	BMessage *msg = new BMessage;
	msg->AddBool(kPrefsWarpingEnabled, m_WarpingEnabled->Value());
	msg->AddInt16(kPrefsWarpingDelay, (int16)m_WarpingDelay->Value());
	msg->AddBool(kPrefsPopupEnabled, m_PopupEnabled->Value());
	msg->AddBool(kPrefsPopupAnimationEnabled, m_PopupAnimationEnabled->Value());
	msg->AddInt16(kPrefsPopupSize, (int16)m_PopupSize->Value());
	msg->AddInt8(kPrefsPopupPosition, m_PopupPosition->ActiveButton());
	msg->AddInt8(kPrefsPopupDirection, m_PopupDirection->ActiveButton());
	msg->AddInt32(kPrefsWarpingIndexField, 0);

	int32 count = 32;
	g_AppPrefs->RemoveName(kPrefsWarpingMsg);
	for (int32 i=0; i<count; i++) {
		msg->ReplaceInt32(kPrefsWarpingIndexField, i);
		g_AppPrefs->AddMessage(kPrefsWarpingMsg, msg);
	}

	delete msg;
	
	BroadcastChanges();
}

